<?php
declare(strict_types = 1);

namespace ctr\api;

use lib\Captcha;
use lib\Core;
use lib\Db;
use lib\Fs;
use mod\Account;
use mod\File;
use mod\Folder;
use mod\Mod;
use mod\Setting;
use mod\Vfile;
use sys\Ctr;

class Zuny extends Ctr {

    public function onLoad() {
        $db = Db::get();
        if (!$db->connect()) {
            return 'Site has an error.';
        }
        Mod::setDb($db);
        $this->_startSession($db);
    }

    // --- 获取验证码 ---
    public function captcha() {
        $cap = Captcha::get(240, 90);
        $_SESSION['captcha'] = [
            'code' => $cap->getPhrase(),
            'exp' => $_SERVER['REQUEST_TIME'] + 120
        ];
        return $cap->getBuffer();
    }

    // --- 获取登录状态 ---
    public function checkLogin() {
        $acc = Account::getLogined();
        return [1, 'account' => $acc ? $acc->name : false];
    }

    // --- 登录 ---
    public function login() {
        // --- 加载语言 ---
        $setting = Setting::where([
            'name' => ['lang']
        ])->all('name');
        $this->_loadLocale($setting['lang']->value);
        if (!$this->_checkXInput($this->_post, [
            'name' => ['require', [0, l('name-can-not-be-empty')]],
            'password' => ['require', [0, l('password-can-not-be-empty')]],
            'captcha' => ['require', [0, l('captcha-can-not-be-empty')]]
        ], $rtn)) {
            return $rtn;
        }
        if (!isset($_SESSION['captcha'])) {
            return [0, 'Param error.'];
        }
        // --- 判断验证码 ---
        if ($_SESSION['captcha']['exp'] < $_SERVER['REQUEST_TIME']) {
            unset($_SESSION['captcha']);
            return [0, l('the-captcha-has-expired')];
        }
        if (strtolower($this->_post['captcha']) !== strtolower($_SESSION['captcha']['code'])) {
            unset($_SESSION['captcha']);
            return [0, l('the-captcha-is-invalid')];
        }
        unset($_SESSION['captcha']);
        $acc = Account::one([
            'name' => $this->_post['name']
        ]);
        if (!$acc) {
            return [0, l('the-account-or-password-is-incorrect')];
        }
        if ($acc->password !== hash_hmac('md5', $this->_post['password'], $acc->salt)) {
            return [0, l('the-account-or-password-is-incorrect')];
        }
        if ((int)$acc->ban) {
            return [0, l('your account has been disabled')];
        }
        if (EDITION_NUM === EDITION_TRIAL_NUM) {
            if (!$acc->checkPermission('root')) {
                return [0, l('in the current version, only root users can log in')];
            }
        }
        $_SESSION['user'] = [
            'id' => $acc->id,
            'salt' => $acc->salt
        ];
        return [1, 'account' => $acc->name];
    }

    // --- 登出 ---
    public function logout() {
        unset($_SESSION['user']);
        return [1];
    }

    // --- 创建文件夹 ---
    public function createFolder() {
        // --- 加载语言 ---
        $setting = Setting::where([
            'name' => ['lang', 'fsplit']
        ])->all('name');
        $this->_loadLocale($setting['lang']->value);
        if (!$this->_checkXInput($this->_post, [
            'name' => ['require', [0, l('name-can-not-be-empty')]],
            'note' => []
        ], $rtn)) {
            return $rtn;
        }
        if (!isset($this->_post['parent'])) {
            $this->_post['parent'] = '';
        }
        $acc = Account::getLogined();
        if (!$acc) {
            return [0, l('please-log-in-first')];
        }
        if (!$acc->checkPermission('root')) {
            return [0, l('no-permission')];
        }
        // --- 判断分表 ---
        $index = '0';
        if ($setting['fsplit']->value > 1) {
            // --- 分表 ---
        }
        $rtn = Folder::createFolder($this->_post['name'], $this->_post['note'], $this->_post['parent'], '1', $acc->id, $index);
        if ($rtn === null) {
            return [0, 'Param error.'];
        }
        if (!$rtn) {
            return [0, l('cannot-duplicate-names')];
        }
        return [1];
    }

    // --- 上传文件 ---
    public function upload() {
        if (EDITION_NUM > 1) {
            return [0, 'Param error.'];
        }
        if (!$this->_checkXInput($this->_post, [
            'parent' => []
        ], $rtn)) {
            return $rtn;
        }
        $path = '';
        $pathKey = '';
        if ($this->_post['parent']) {
            // --- $parent 是上层，有可能为空，空代表的顶级 ---
            $pf = Folder::one([
                'key' => $this->_post['parent']
            ], false, '0');
            if (!$pf) {
                return [0, 'Param error.'];
            }
            if ($pf->path) {
                $path = $pf->path . '/';
                $pathKey = $pf->path_key . '/';
            }
            $path .= $pf->name;
            $pathKey .= $pf->key;
        }
        // --- upload 为上传模式，空为上传到本地 ---
        $setting = Setting::where([
            'name' => ['lang', 'upload', 'fsplit', 'vsplit']
        ])->all('name');
        $this->_loadLocale($setting['lang']->value);
        $acc = Account::getLogined();
        if (!$acc) {
            return [0, l('please-log-in-first')];
        }
        if (!$acc->checkPermission('root')) {
            return [0, l('no-permission')];
        }
        if (!isset($_FILES['file'])) {
            return [0, 'Param error.'];
        }
        if (!$acc->requestSpace($_FILES['file']['size'])) {
            return [0, l('lack-of-space')];
        }
        // --- 只会上传到本地 ---
        /** --- 相对路径 --- */
        $rpath = date('Y/m/d/');
        /** --- 实体文件名 --- */
        $fname = Core::muid([
            'len' => '32',
            'bin' => false
        ]) . '.mnsvf';
        /** --- 绝对路径 --- */
        $fpath = DATA_PATH . 'files/' . $rpath;
        if (!Fs::mkdir($fpath)) {
            $acc->restoreSpace($_FILES['file']['size']);
            return [0, l('upload-failed')];
        }
        if (!move_uploaded_file($_FILES['file']['tmp_name'], $fpath . $fname)) {
            $acc->restoreSpace($_FILES['file']['size']);
            return [0, l('upload-failed')];
        }
        // --- 创建相关数据库信息 ---
        // --- 判断分表 ---
        $vindex = '0';
        if ($setting['vsplit']->value > 1) {
            // --- 分表 ---
        }
        $findex = '0';
        if ($setting['fsplit']->value > 1) {
            // --- 分表 ---
        }
        $time = time();
        $vfileKey = Core::muid([
            'len' => '40'
        ]);
        $vfile = Vfile::getCreate($vindex);
        $vfile->set([
            'key' => $vfileKey,
            'size' => $_FILES['file']['size'],
            'sha256' => '',
            'server' => '0',
            'path' => $rpath . $fname,
            'ref' => '1',
            'time_add' => $time
        ]);
        $vfile->create();
        // --- 将文件名分离 ---
        $fsp = strrpos($_FILES['file']['name'], '.');
        $fname = $fsp === false ? $_FILES['file']['name'] : substr($_FILES['file']['name'], 0, $fsp);
        $ext = $fsp === false ? '' : substr($_FILES['file']['name'], $fsp + 1);
        // --- 检测文件名不能重复 ---
        $count = 0;
        while (true) {
            if (!Folder::checkDuplicate($fname . ($count ? $count : '') . ($ext ? '.' . $ext : ''), $this->_post['parent'], '1', $findex)) {
                break;
            }
            ++$count;
        }
        if ($count) {
            $fname = $fname . $count;
        }
        $file = File::getCreate($findex);
        $file->set([
            'key' => Core::muid([
                'len' => '40'
            ]),
            'owner' => '1',
            'creator' => $acc->id,
            'parent' => $this->_post['parent'],
            'name' => $fname . ($ext ? '.' . $ext : ''),
            'ext' => $ext,
            'size' => $_FILES['file']['size'],
            'path' => $path,
            'path_key' => $pathKey,
            'down' => '0',
            'visit' => '0',
            'vfile' => $vfileKey,
            'url' => '',
            'time_add' => $time,
            'time_edit' => $time
        ]);
        $file->create();
        Account::updateByWhere([
            ['files', '+', '1']
        ], [
            'id' => '1'
        ]);
        if ($this->_post['parent']) {
            $parent = $this->_post['parent'];
            while ($parent) {
                $pf = Folder::one([
                    'key' => $parent
                ], false, $findex);
                Folder::updateByWhere([
                    ['files', '+', '1'],
                    ['size', '+', $_FILES['file']['size']]
                ], [
                    'id' => $pf->id
                ], false, $findex);
                $parent = $pf->parent;
            }
        }
        return [1];
    }

    // --- 移动 ---
    public function move() {
        if (!$this->_checkXInput($this->_post, [
            'key' => ['require', [0, 'Key can not be empty.']],
            'type' => ['require', ['folder', 'file'], [0, 'Type can not be empty.']],
            'to' => []
        ], $rtn)) {
            return $rtn;
        }
        $setting = Setting::where([
            'name' => ['lang', 'fsplit']
        ])->all('name');
        $this->_loadLocale($setting['lang']->value);
        $acc = Account::getLogined();
        if (!$acc) {
            return [0, l('please-log-in-first')];
        }
        if (!$acc->checkPermission('root')) {
            return [0, l('no-permission')];
        }
        // --- 判断分表 ---
        $index = '0';
        if ($setting['fsplit']->value > 1) {
            // --- 分表 ---
        }
        if ($this->_post['type'] === 'file') {
            // --- 文件 ---
            $rtn = File::moveFile($this->_post['key'], $this->_post['to'], $index);
            if ($rtn <= 0) {
                switch ($rtn) {
                    case 0: {
                        return [0, l('file-not-exist')];
                    }
                    case -1: {
                        return [0, l('cannot-duplicate-names')];
                    }
                }
            }
        }
        else {
            // --- 文件夹 ---
            $rtn = Folder::moveFolder($this->_post['key'], $this->_post['to'], $index);
            if ($rtn <= 0) {
                switch ($rtn) {
                    case 0: {
                        return [0, l('folder-not-exist')];
                    }
                    case -1: {
                        return [0, l('cannot-duplicate-names')];
                    }
                    case -2: {
                        return [0, l('cannot move to own subdirectory')];
                    }
                }
            }
        }
        return [1];
    }

    // --- 移除 ---
    public function remove() {
        if (!$this->_checkXInput($this->_post, [
            'key' => ['require', [0, 'Key can not be empty.']],
            'type' => ['require', ['folder', 'file'], [0, 'Type can not be empty.']]
        ], $rtn)) {
            return $rtn;
        }
        // --- 加载语言 ---
        $setting = Setting::where([
            'name' => ['lang', 'fsplit', 'vsplit']
        ])->all('name');
        $this->_loadLocale($setting['lang']->value);
        $acc = Account::getLogined();
        if (!$acc) {
            return [0, l('please-log-in-first')];
        }
        if (!$acc->checkPermission('root')) {
            return [0, l('no-permission')];
        }
        // --- 判断分表 ---
        $findex = '0';
        if ($setting['fsplit']->value > 1) {
            // --- 分表 ---
        }
        $vsplit = $setting['vsplit']->value;
        if ($this->_post['type'] === 'folder') {
            // --- 文件夹 ---
            $folder = Folder::one([
                'key' => $this->_post['key']
            ], false, $findex);
            if (!$folder) {
                return [0, l('folder-not-exist')];
            }
            $folder->delete($vsplit);
            return [1];
        }
        else {
            // --- 文件 ---
            $file = File::one([
                'key' => $this->_post['key']
            ], false, $findex);
            if (!$file) {
                return [0, l('file-not-exist')];
            }
            $file->delete($vsplit);
            return [1];
        }
    }

    // --- 重命名文件/文件夹 ---
    public function rename() {
        // --- 加载语言 ---
        $setting = Setting::where([
            'name' => ['lang', 'fsplit']
        ])->all('name');
        $this->_loadLocale($setting['lang']->value);
        if (!$this->_checkXInput($this->_post, [
            'key' => ['require', [0, 'Key can not be empty.']],
            'name' => ['require', [0, l('name-can-not-be-empty')]],
            'note' => [],
            'type' => ['require', ['folder', 'file'], [0, 'Type can not be empty.']]
        ], $rtn)) {
            return $rtn;
        }
        $acc = Account::getLogined();
        if (!$acc) {
            return [0, l('please-log-in-first')];
        }
        if (!$acc->checkPermission('root')) {
            return [0, l('no-permission')];
        }
        // --- 判断分表 ---
        $index = '0';
        if ($setting['fsplit']->value > 1) {
            // --- 分表 ---
        }
        if ($this->_post['type'] === 'folder') {
            // --- 文件夹 ---
            $folder = Folder::one([
                'key' => $this->_post['key']
            ], false, $index);
            if (!$folder) {
                return [0, l('folder-not-exist')];
            }
            if (!$folder->rename($this->_post['name'], $this->_post['note'])) {
                return [0, l('cannot-have-the-same-name-as-another-file-or-folder')];
            }
            return [1];
        }
        else {
            // --- 文件 ---
            $file = File::one([
                'key' => $this->_post['key']
            ], false, $index);
            if (!$file) {
                return [0, l('file-not-exist')];
            }
            if (!$file->rename($this->_post['name'])) {
                return [0, l('cannot-have-the-same-name-as-another-file-or-folder')];
            }
            return [1];
        }
    }

    // --- 获取列表 ---
    public function getList() {
        $setting = Setting::where([
            'name' => ['open', 'closeinfo', 'title', 'fsplit']
        ])->all('name');
        if ((int)$setting['open']->value !== 1) {
            return [0, $setting['closeinfo']->value];
        }
        // --- 判断分表 ---
        $index = '0';
        if ($setting['fsplit']->value > 1) {
            // --- 分表 ---
        }
        // --- 获取 parent 信息 ---
        $parent = isset($this->_post['parent']) ? $this->_post['parent'] : '';
        $rtn = Folder::getAllList($parent, '1', $index, (isset($this->_post['file']) && ($this->_post['file'] === '0')) ? false : true);
        if (!$rtn) {
            return [0, 'Not found.'];
        }
        return [1,
            'parent' => $parent,
            'title' => ($rtn['name'] ? $rtn['name'] . ' - ' : '') . $setting['title']->value,
            'path' => $rtn['path'],
            'pathKey' => $rtn['pathKey'],
            'list' => $rtn['list']
        ];
    }

}

