<?php
declare(strict_types = 1);

namespace ctr\install;

use lib\Core;
use lib\Db;
use lib\Fs;
use lib\Net;
use mod\Account;
use mod\Mod;
use mod\Setting;
use sys\Ctr;

class Trial extends Ctr {

    // --- 获取 license ---
    public function license() {
        if (is_file(DATA_PATH . 'mns/install.lock')) {
            return [0, 'Param error.'];
        }
        if (!$this->_checkInput($this->_post, [
            'lang' => ['require', [0, 'Param error.']]
        ], $rtn)) {
            return $rtn;
        }
        $fpath = DATA_PATH . 'mns/license.' . $this->_post['lang'] . '.html';
        if (!is_file($fpath)) {
            return [0, 'Param error.'];
        }
        $license = file_get_contents($fpath);
        $license = str_replace('{PRODUCT}', PRODUCT_W, $license);
        return [1, 'license' => $license];
    }

    // --- 检查权限和函数 ---
    public function permission() {
        if (is_file(DATA_PATH . 'mns/install.lock')) {
            return [0, 'Param error.'];
        }
        $rtn = [
            'folder' => [
                'data/mns/' => false,
                'data/files/' => false,
                'data/db/' => false,
                'data/db/db.sqlite' => false,
                'etc/' => false,
                'etc/db.php' => false,
                'etc/kv.php' => false,
                'etc/sql.php' => false,
                'log/' => false,
                'stc/files/' => false
            ],
            'function' => [
                'curl_init' => false,
                'imagecreatefromjpeg' => false,
                'imagecreatefrompng' => false,
                'imagecreatefromgif' => false
            ],
            'class' => [
                'PDO' => false
            ],
            'other' => [
                'PHP 8.0+' => false
            ],
            'existdir' => [
            ],
            'existfile' => [
            ]
        ];
        foreach ($rtn as $type => $item) {
            switch ($type) {
                case 'folder': {
                    foreach ($item as $path => $val) {
                        if (!Fs::isWritable(ROOT_PATH . $path)) {
                            continue;
                        }
                        $rtn[$type][$path] = true;
                    }
                    break;
                }
                case 'function': {
                    foreach ($item as $name => $val) {
                        if (!function_exists($name)) {
                            continue;
                        }
                        $rtn[$type][$name] = true;
                    }
                    break;
                }
                case 'class': {
                    foreach ($item as $name => $val) {
                        if (!class_exists($name)) {
                            continue;
                        }
                        $rtn[$type][$name] = true;
                    }
                    break;
                }
                case 'existdir': {
                    foreach ($item as $name => $val) {
                        if (!is_dir($name)) {
                            continue;
                        }
                        $rtn[$type][$name] = true;
                    }
                    break;
                }
                case 'existfile': {
                    foreach ($item as $name => $val) {
                        if (!is_file($name)) {
                            continue;
                        }
                        $rtn[$type][$name] = true;
                    }
                    break;
                }
                case 'other': {
                    foreach ($item as $name => $val) {
                        if (!version_compare(PHP_VERSION, '8.0.0', '>=')) {
                            continue;
                        }
                        $rtn[$type][$name] = true;
                    }
                    break;
                }
            }
        }
        return [1, $rtn];
    }

    // --- 数据库与缓存信息 ---
    public function db() {
        if (is_file(DATA_PATH . 'mns/install.lock')) {
            return [0, 'Param error.'];
        }
        if (EDITION_NUM > EDITION_TRIAL_NUM) {
            return $this->_location(URL_BASE . 'install/standard/db');
        }
        // --- SQLITE BEGIN ---
        if (!is_file(DATA_PATH . 'db/db.sqlite')) {
            return [0, '"data/db/db.sqlite" does not exist.'];
        }
        $db = Db::get(Db::SQLITE);
        if (!$db->connect()) {
            return [0, 'Connection failed.'];
        }
        // --- SQLITE END ---
        // --- 表前缀 BEGIN ---
        $sqlEtc = file_get_contents(ETC_PATH . 'sql.php');
        $sqlEtc = preg_replace('/SQL_PRE = \'.+?\'/', 'SQL_PRE = \'' . $this->_post['pre'] . '\'', $sqlEtc);
        file_put_contents(ETC_PATH . 'sql.php', $sqlEtc);
        // --- 表前缀 END ---
        return [1];
    }

    // --- 准备好 ---
    public function ready() {
        if (is_file(DATA_PATH . 'mns/install.lock')) {
            return [0, 'Param error.'];
        }
        // --- 判断数据库类型 ---
        $dbEtc = file_get_contents(ETC_PATH . 'db.php');
        $isMysql = true;
        if (strpos($dbEtc, 'MY_HOST = \'\'')) {
            $isMysql = false;
        }
        // --- 获取要导入的 sql 文件 ---
        $sql = file_get_contents(DATA_PATH . 'db/' . ($isMysql ? 'mysql' : 'sqlite') . '.sql');
        if (!$sql) {
            return [0, 'Param error.'];
        }
        require ETC_PATH.'sql.php';
        $sql = str_replace('\r\n', '\n', $sql);
        $sql = str_replace('\r', '\n', $sql);
        $sql = preg_replace('/-- .+?\n/', '', $sql);
        $sql = str_replace('mns_', SQL_PRE, $sql);
        $sqls = [];
        $arr = explode(';', $sql);
        foreach ($arr as $line) {
            $line = trim($line);
            if (!$line) {
                continue;
            }
            $sqls[] = $line;
        }
        return [1, 'sqls' => $sqls];
    }

    // --- 安装 ---
    public function install() {
        if (is_file(DATA_PATH . 'mns/install.lock')) {
            return [0, 'Param error.'];
        }
        $db = Db::get();
        if (!$db->connect()) {
            return [0, 'Connection failed.'];
        }
        $r = $db->exec($this->_post['sql']);
        if ($r === false) {
            return [0, 'Exec failed (' . $db->getErrorInfo()[2] . ').'];
        }
        if ($r < 0) {
            return [0, 'Exec failed.'];
        }
        return [1];
    }

    // --- 设置站点信息 ---
    public function setting() {
        if (is_file(DATA_PATH . 'mns/install.lock')) {
            return [0, 'Param error.'];
        }
        if (!$this->_checkInput($this->_post, [
            'name' => ['require', [0, 'Name require']],
            'password' => ['require', [0, 'Password require']],
            'email' => ['/^.+?@.+?$/', [0, 'Email error']],
            'area' => ['num', [0, 'Area error']],
            'phone' => ['num', [0, 'Phone error']],
            'title' => ['require', [0, 'Site name require']],
            'lang' => ['require', [0, 'Param lang error']]
        ], $rtn)) {
            return $rtn;
        }
        $db = Db::get();
        if (!$db->connect()) {
            return [0, 'Connection failed.'];
        }
        Mod::setDb($db);
        $salt = Core::random(4);
        $account = Account::getCreate();
        $account->set([
            'name' => $this->_post['name'],
            'password' => hash_hmac('md5', $this->_post['password'], $salt),
            'salt' => $salt,
            'email' => $this->_post['email'] ? $this->_post['email'] : '',
            'area' => $this->_post['area'] ? $this->_post['area'] : '0',
            'phone' => $this->_post['phone'] ? $this->_post['phone'] : '0',
            'permission' => 'root',
            'disk_usage' => '0',
            'disk_max' => '1099511627776',
            'disk_remain' => '1099511627776',
            'files' => '0',
            'folders' => '0',
            'ban' => '0',
            'time_add' => time()
        ]);
        if (!$account->create()) {
            return [0, 'Account create failed.'];
        }
        if (Setting::updateByWhere([
            'value' => $this->_post['title']
        ], [
            'name' => 'title'
        ]) === false) {
            return [0, 'Site name update failed.'];
        }
        if (Setting::updateByWhere([
            'value' => $this->_post['lang']
        ], [
            'name' => 'lang'
        ]) === false) {
            return [0, 'Lang update failed.'];
        }
        return [1];
    }

    // --- 体验计划，最后一步 ---
    public function improvement() {
        if (is_file(DATA_PATH . 'mns/install.lock')) {
            return [0, 'Param error.'];
        }
        if (!$this->_checkInput($this->_post, [
            'email' => ['/^.+?@.+?$/', [0, 'Email error']],
            'area' => ['num', [0, 'Area error']],
            'phone' => ['num', [0, 'Phone error']],
            'qq' => ['num', [0, 'QQ error']]
        ], $rtn)) {
            return $rtn;
        }
        // --- 提交信息，在中国大陆境内的服务器提交至上海，除中国大陆以外的地区提交至新加坡 ---
        if (!$this->_post['phone'] && !$this->_post['email']) {
            $this->_post['email'] = 'none@email.com';
        }
        $xsrf = Core::random(8);
        $cookie = [];
        Net::setCookie($cookie, 'XSRF-TOKEN', $xsrf, 'www.maiyun.net');
        Net::postJson('https://www.maiyun.net/api/issue', [
            'type' => '5',
            'mcode' => 'cny',
            'budget' => '0',
            'product' => PRODUCT_S,
            'content' => 'QQ' . ($this->_post['qq'] ? $this->_post['qq'] : '') . ',' . URL_FULL . ',' . $_SERVER['HTTP_USER_AGENT'],
            'uarea' => $this->_post['area'] ? $this->_post['area'] : '0',
            'phone' => $this->_post['phone'],
            'email' => $this->_post['email'],
            'lang' => $this->_post['lang'],
            '_xsrf' => $xsrf
        ], [], $cookie);
        @file_put_contents(DATA_PATH . 'mns/install.lock', 'mns.lock.' . PRODUCT);
        if (!is_file(DATA_PATH . 'mns/install.lock')) {
            return [0, 'Failed.'];
        }
        return [1];
    }

}

