<?php
declare(strict_types = 1);

namespace mod;

use lib\Consistent;
use lib\Core;
use lib\Text;

class Folder extends Mod {

    protected static $_table = 'folder';
    protected static $_primary = 'id';

    public $id, $key, $owner, $creator, $parent, $name, $note, $size, $files, $folders, $path, $path_key, $time_add, $time_edit;

    /**
     * --- 将文件夹更名为新名 ---
     * @param $name 新名
     */
    public function rename(string $name, string $note): bool {
        if (!$name) {
            return false;
        }
        if ($name === $this->name) {
            // --- 只修改 note ---
            if (!Folder::updateByWhere([
                'note' => $note,
                'time_edit' => time()
            ], [
                'key' => $this->key
            ], false, $this->_index)) {
                return false;
            }
            $this->note = $note;
            return true;
        }
        // --- 改名和 note ---
        if (Folder::checkDuplicate($name, $this->parent, $this->owner, $this->_index)) {
            return false;
        }
        if (!Folder::updateByWhere([
            'name' => $name,
            'note' => $note,
            'time_edit' => time()
        ], [
            'key' => $this->key
        ], false, $this->_index)) {
            return false;
        }
        $this->name = $name;
        $this->note = $note;
        return true;
    }

    /**
     * --- 删除当前文件夹 ---
     */
    public function delete($vHashCount = 1, $raw = false, $level = 0): int {
        $nodes = null;
        $files = File::where([
            'owner' => $this->owner,
            'parent' => $this->key
        ], false, $this->_index)->all();
        foreach ($files as $file) {
            $file->remove($raw);
            if ($file->vfile) {
                $index = '0';
                if ($vHashCount > 1) {
                    if (!$nodes) {
                        $nodes = Consistent::getRange(0, $vHashCount - 1);
                    }
                    $index = Consistent::fast($file->vfile, $nodes);
                }
                Vfile::updateByWhere([
                    ['ref', '-', '1']
                ], [
                    'key' => $file->vfile
                ], false, $index);
            }
        }
        $folders = Folder::where([
            'owner' => $this->owner,
            'parent' => $this->key
        ], $raw, $this->_index)->all();
        foreach ($folders as $folder) {
            $folder->delete($vHashCount, $raw, $level + 1);
        }
        $this->remove($raw);
        if (!$level) {
            if ($this->parent) {
                Folder::updateByWhere([
                    ['files', '-', $this->files],
                    ['folders', '-', $this->folders + 1],
                    ['size', '-', $this->size]
                ], [
                    'key' => $this->parent
                ], $raw, $this->_index);
            }
            Account::updateByWhere([
                ['files', '-', $this->files],
                ['folders', '-', $this->folders + 1],
                ['disk_usage', '-', $this->size],
                ['disk_remain', '+', $this->size]
            ], [
                'id' => $this->owner
            ]);
        }
        return $this->size;
    }

    /** --- 判断此层级有没有重名的文件、文件夹，不检查 parent 是否存在，返回 true 代表重名了 --- */
    public static function checkDuplicate($name, $parent, $owner, $index): bool {
        if (strpos($name, '/') !== false) {
            return true;
        }
        if (Folder::where([
            'owner' => $owner,
            'parent' => $parent,
            'name' => $name
        ], false, $index)->count() > 0) {
            return true;
        }
        if (File::where([
            'owner' => $owner,
            'parent' => $parent,
            'name' => $name
        ], false, $index)->count() > 0) {
            return true;
        }
        return false;
    }

    /** --- 创建文件夹 --- */
    public static function createFolder($name, $note, $parent, $owner, $creator, $index): bool {
        $path = '';
        $pathKey = '';
        if ($parent) {
            // --- $parent 是要创建文件夹的上层，有可能为空，空代表的顶级 ---
            $pf = Folder::one([
                'key' => $parent
            ], false, $index);
            if (!$pf) {
                return null;
            }
            if ($pf->path) {
                $path = $pf->path . '/';
                $pathKey = $pf->path_key . '/';
            }
            $path .= $pf->name;
            $pathKey .= $pf->key;
        }
        if (Folder::checkDuplicate($name, $parent, $owner, $index)) {
            return false;
        }
        $folder = Folder::getCreate($index);
        $folder->set([
            'key' => Core::muid([
                'len' => 40
            ]),
            'owner' => $owner,
            'creator' => $creator,
            'parent' => $parent,
            'name' => $name,
            'note' => $note,
            'size' => '0',
            'files' => '0',
            'folders' => '0',
            'path' => $path,
            'path_key' => $pathKey,
            'time_add' => $_SERVER['REQUEST_TIME'],
            'time_edit' => $_SERVER['REQUEST_TIME']
        ]);
        if (!$folder->create()) {
            return false;
        }
        if ($parent) {
            while ($parent) {
                $pf = Folder::one([
                    'key' => $parent
                ], false, $index);
                Folder::updateByWhere([
                    ['folders', '+', '1']
                ], [
                    'id' => $pf->id
                ], false, '0');
                $parent = $pf->parent;
            }
        }
        Account::updateByWhere([
            ['folders', '+', '1']
        ], [
            'id' => $owner
        ]);
        return true;
    }

    /**
     * --- 移动一个文件夹到一个新位置 ---
     */
    public static function moveFolder($key, $parent, $index): int {
        $folder = Folder::one([
            'key' => $key
        ], false, $index);
        if (!$folder) {
            return 0;
        }
        if ($folder->parent == $parent) {
            return 1;
        }
        if (Folder::checkDuplicate($folder->name, $parent, $folder->owner, $index)) {
            return -1;
        }
        $newPath = '';
        $newPathKey = '';
        if ($parent) {
            // --- $parent 是要移动到的目标文件夹，有可能为空，空代表的顶级 ---
            $pf = Folder::one([
                'key' => $parent
            ], false, $index);
            if (!$pf) {
                return 0;
            }
            if ($pf->path) {
                $newPath = $pf->path . '/';
                $newPathKey = $pf->path_key . '/';
            }
            $newPath .= $pf->name;
            $newPathKey .= $pf->key;
        }
        // --- 判断别把自己移动到自己的子文件夹里头去了 ---
        if (strpos('/' . $newPathKey . '/', '/' . $folder->key . '/') !== false) {
            return -2;
        }
        // --- 移动 ---
        if (!Folder::updateByWhere([
            'path' => $newPath,
            'path_key' => $newPathKey,
            'parent' => $parent
        ], [
            'id' => $folder->id
        ], false, $index)) {
            return 0;
        }
        // --- 给老的 parent 减少 folders、size、files 等计数 ---
        $oldParent = $folder->parent;
        if ($oldParent) {
            while ($oldParent) {
                $pf = Folder::one([
                    'key' => $oldParent
                ], false, $index);
                Folder::updateByWhere([
                    ['folders', '-', ($folder->folders + 1)],
                    ['files', '-', $folder->files],
                    ['size', '-', $folder->size]
                ], [
                    'id' => $pf->id
                ], false, $index);
                if ($pf->parent === '') {
                    break;
                }
                $oldParent = $pf->parent;
            }
        }
        // --- 给新的增加上面说的那些 ---
        if ($parent) {
            while ($parent) {
                $pf = Folder::one([
                    'key' => $parent
                ], false, $index);
                Folder::updateByWhere([
                    ['folders', '+', ($folder->folders + 1)],
                    ['files', '+', $folder->files],
                    ['size', '+', $folder->size]
                ], [
                    'id' => $pf->id
                ], false, $index);
                if ($pf->parent === '') {
                    break;
                }
                $parent = $pf->parent;
            }
        }
        return 1;
    }

    /** --- 获取所有子文件夹文件，parent 不存在则返回 null --- */
    public static function getAllList($parent, $owner, $index, $file = true): array | null {
        $rtn = [
            'name' => '',
            'path' => [],
            'pathKey' => [],
            'list' => []
        ];
        // --- 地址 ---
        if ($parent) {
            $folder = Folder::one([
                'key' => $parent
            ], false, $index);
            if (!$folder) {
                return null;
            }
            $rtn['name'] = $folder->name;
            if ($folder->path) {
                $rtn['path'] = explode('/', $folder->path);
                $rtn['pathKey'] = explode('/', $folder->path_key);
            }
            $rtn['path'][] = $folder->name;
            $rtn['pathKey'][] = $folder->key;
        }
        // --- 获取列表 ---
        $folders = Folder::where([
            'owner' => $owner,
            'parent' => $parent
        ], false, $index)->by('time_edit')->all();
        foreach ($folders as $item) {
            $rtn['list'][] = [
                'type' => 'folder',
                'key' => $item->key,
                'name' => $item->name,
                'note' => $item->note,
                'size' => Text::sizeFormat((int)$item->size),
                'time_add' => date('Y-m-d H:i', (int)$item->time_add),
                'time_edit' => date('Y-m-d H:i', (int)$item->time_edit),
            ];
        }
        if ($file) {
            $files = File::where([
                'owner' => $owner,
                'parent' => $parent
            ], false, $index)->by('time_edit')->all();
            foreach ($files as $item) {
                $rtn['list'][] = [
                    'type' => 'file',
                    'key' => $item->key,
                    'name' => $item->name,
                    'size' => Text::sizeFormat((int)$item->size),
                    'time_add' => date('Y-m-d H:i', (int)$item->time_add),
                    'time_edit' => date('Y-m-d H:i', (int)$item->time_edit),
                ];
            }
        }
        return $rtn;
    }

}

